import { AnnotationRange } from 'shared-react/annotation/selection.model';
import { default as default_2 } from 'react';
import { immutableNoteCallerNodeName } from 'shared-react/nodes/scripture/usj/ImmutableNoteCallerNode';
import { LoggerBasic } from 'shared-react/plugins/logger-basic.model';
import { ScriptureReference } from 'platform-bible-utils';
import { SelectionRange } from 'shared-react/annotation/selection.model';
import { TextDirection } from 'shared-react/plugins/text-direction.model';
import { Usj } from '../../../utilities/src';
import { UsjNodeOptions } from 'shared-react/nodes/scripture/usj/usj-node-options.model';

declare type Comment_2 = {
    author: string;
    content: string;
    deleted: boolean;
    id: string;
    timeStamp: number;
    type: "comment";
};

export declare type Comments = Array<Thread | Comment_2>;

export declare const DEFAULT_VIEW_MODE = "formatted";

/**
 * Scripture Editor for USJ. Created for use in [Platform](https://platform.bible).
 * @see https://github.com/usfm-bible/tcdocs/blob/usj/grammar/usj.js
 *
 * @param props.ref - Forward reference for the editor.
 * @param props.defaultUsj - Initial Scripture data in USJ format.
 * @param props.scrRef - Scripture reference that links the general cursor location of the
 *   Scripture.
 * @param props.onScrRefChange - Callback function when the Scripture reference changes in the editor as
 *   the cursor moves.
 * @param props.options - Options to configure the editor.
 * @param props.onUsjChange - Callback function when USJ Scripture data has changed.
 * @param props.logger - Logger instance.
 * @returns the editor element.
 */
export declare const Editorial: default_2.ForwardRefExoticComponent<EditorProps<LoggerBasic> & default_2.RefAttributes<EditorRef | null>>;

/** Options to configure the editor. */
export declare type EditorOptions = {
    /** Is the editor readonly or editable. */
    isReadonly?: boolean;
    /** Is the editor enabled for spell checking. */
    hasSpellCheck?: boolean;
    /** Text direction: "ltr" | "rtl" | "auto". */
    textDirection?: TextDirection;
    /**
     * View options - EXPERIMENTAL. Defaults to the formatted view mode which is currently the only
     * functional option.
     */
    view?: ViewOptions;
    /** Options for each editor node:
     * @param nodes.ImmutableNoteCallerNode.noteCallers - Possible note callers to use when caller is
     *   '+'. Defaults to Latin lower case letters.
     * @param nodes.ImmutableNoteCallerNode.onClick - Click handler method.
     */
    nodes?: UsjNodeOptions;
};

declare type EditorProps<TLogger extends LoggerBasic> = {
    /** Initial Scripture data in USJ format. */
    defaultUsj?: Usj;
    /** Scripture reference that controls the general cursor location of the Scripture. */
    scrRef?: ScriptureReference;
    /** Callback function when the Scripture reference has changed. */
    onScrRefChange?: (scrRef: ScriptureReference) => void;
    /** Options to configure the editor. */
    options?: EditorOptions;
    /** Callback function when USJ Scripture data has changed. */
    onUsjChange?: (usj: Usj) => void;
    /** Logger instance. */
    logger?: TLogger;
};

/** Forward reference for the editor. */
export declare type EditorRef = {
    /** Focus the editor. */
    focus(): void;
    /** Set the USJ Scripture data. */
    setUsj(usj: Usj): void;
    /**
     * Get the selection location or range.
     * @returns the selection location or range, or `undefined` if there is no selection. The
     *   json-path in the selection assumes no comment Milestone nodes are present in the USJ.
     */
    getSelection(): SelectionRange | undefined;
    /**
     * Set the selection location or range.
     * @param selection - A selection location or range. The json-path in the selection assumes no
     *   comment Milestone nodes are present in the USJ.
     */
    setSelection(selection: SelectionRange): void;
    /**
     * Add an ephemeral annotation.
     * @param selection - An annotation range containing the start and end location. The json-path in
     *   an annotation location assumes no comment Milestone nodes are present in the USJ.
     * @param type - Type of the annotation.
     * @param id - ID of the annotation.
     */
    addAnnotation(selection: AnnotationRange, type: string, id: string): void;
    /**
     * Remove an ephemeral annotation.
     * @param type - Type of the annotation.
     * @param id - ID of the annotation.
     */
    removeAnnotation(type: string, id: string): void;
    /** Ref to the end of the toolbar - INTERNAL USE ONLY to dynamically add controls in the toolbar. */
    toolbarEndRef: default_2.RefObject<HTMLElement> | null;
};

/**
 * Get view option properties based on the view mode.
 * @param viewMode - View mode of the editor.
 * @returns the view options if the view exists, the default options if the viewMode is undefined,
 *   `undefined` otherwise.
 */
export declare function getViewOptions(viewMode?: string | undefined): ViewOptions | undefined;

export { immutableNoteCallerNodeName }

/**
 * Scripture Editor for USJ with comments in the margin. Created for use in [Platform](https://platform.bible).
 * @see https://github.com/usfm-bible/tcdocs/blob/usj/grammar/usj.js
 *
 * @param props.ref - Forward reference for the editor.
 * @param props.defaultUsj - Initial Scripture data in USJ format.
 * @param props.scrRef - Scripture reference that links the general cursor location in the
 *   Scripture.
 * @param props.onScrRefChange - Callback function when the Scripture reference changes in the editor as
 *   the cursor moves.
 * @param props.options - Options to configure the editor.
 * @param props.onUsjChange - Callback function when USJ Scripture data has changed.
 * @param props.logger - Logger instance.
 * @returns the editor element.
 */
export declare const Marginal: default_2.ForwardRefExoticComponent<Omit<EditorProps<LoggerBasic>, "onUsjChange"> & {
    /** Callback function when USJ Scripture data has changed. */
    onUsjChange?: ((usj: Usj, comments: Comments | undefined) => void) | undefined;
} & default_2.RefAttributes<MarginalRef>>;

/** Forward reference for the editor. */
export declare type MarginalRef = EditorRef & {
    /** Set the comments to accompany USJ Scripture. */
    setComments?(comments: Comments): void;
};

declare type Thread = {
    comments: Array<Comment_2>;
    id: string;
    quote: string;
    type: "thread";
};

export { UsjNodeOptions }

export declare type ViewMode = ViewNameKey;

declare const viewModeToViewNames: {
    formatted: string;
    unformatted: string;
};

declare type ViewNameKey = keyof typeof viewModeToViewNames;

export declare type ViewOptions = {
    /** USFM markers are visible, editable or hidden */
    markerMode: "visible" | "editable" | "hidden";
    /** does the text have spacing including indenting */
    hasSpacing: boolean;
    /** is the text in a formatted font */
    isFormattedFont: boolean;
};

/**
 * Convert view options to view mode if the view exists.
 * @param viewOptions - View options of the editor.
 * @returns the view mode if the view is defined, `undefined` otherwise.
 */
export declare function viewOptionsToMode(viewOptions: ViewOptions | undefined): ViewMode | undefined;

export { }
